@@ -1,10 +1,9 @@
-# Copyright 2001-15 by Roger Bivand 
+# Copyright 2001-24 by Roger Bivand 
 #
 
-.spdepOptions <- new.env(FALSE, globalenv())
+.spdepOptions <- new.env(TRUE, globalenv())
 assign("spChkID", FALSE, envir = .spdepOptions)
 assign("zeroPolicy", FALSE, envir = .spdepOptions)
-assign("report_nb_subgraphs", FALSE, envir = .spdepOptions)
 assign("verbose", FALSE, envir = .spdepOptions)
 assign("mc", ifelse(.Platform$OS.type == "windows", FALSE, TRUE),
  envir = .spdepOptions)
@@ -12,7 +11,10 @@ assign("cores", NULL, envir = .spdepOptions)
 assign("cluster", NULL, envir = .spdepOptions)
 assign("rlecuyerSeed", rep(12345, 6), envir = .spdepOptions)
 assign("listw_is_CsparseMatrix", FALSE, envir = .spdepOptions)
-
+assign("cluster", NULL, envir = .spdepOptions)
+assign("report_nb_subgraphs", TRUE, envir = .spdepOptions)
+assign("nb_subgraphs_N+E", 100000L, envir = .spdepOptions)
+assign("report_nb_noneighs", TRUE, envir = .spdepOptions)
 setOldClass(c("listw"))
 
 .onLoad <- function(lib, pkg) {
